#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
progname = "ocrpdf-tk.py"
versionnr = "0.2.0"
programname = """
		"""+progname+"""

		GUI Wrapper for OCRmyPDF
		Version """+versionnr

licence = """
Copyright 2019 Dr. Michael Stehmann <info@rechtsanwalt-stehmann.de>

This program is free software; you can redistribute it and/or modify
under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
MA 02110-1301, USA."""

# This program is a GUI wrapper for OCRmyPDF
# In Debian the following packages have to be installed additionally:
# -- python3-tk
# -- ocrmypdf
# -- tesseract-ocr-all
# -- atril

import configparser
from tkinter import *
from tkinter import filedialog
from tkinter import messagebox
from tkinter import scrolledtext
from tkinter import simpledialog
import os
import sys
import locale
import gettext

# def _(string):
#     """
#     Dummy function for gettext
#     """
#     return string

class DialogMaker(object):
	"""Methods to create and design Tkinter dialogs"""

	def __init__(self):
		self.darkbgcolor = "#C7DAE9"
		self.brightcolor = "#F6F9AC"
		"""self.fs contains the font size
		10 seems to be a good default value"""
		self.fs = 10

	def makedia(self):
		"""Generates Tkinter dialog"""
		self.dia=Tk()

	def diatitle (self, title):
		"""Sets dialog title"""
		self.dia.title(title)

	def diaminsize(self, diah=900, diav=450):
		"""Minimum size of the dialog window
		diah = horizontal minimum size
		diav = vertical minimum size"""
		self.dia.minsize(diah, diav)

	def geometry(self, geostring="+50+0"):
		"""Places dialog on the desktop,
		the first number is horizontal, the second vertical"""
		self.dia.geometry(geostring)

	def mainloop(self):
		"""Creates Dialog"""
		self.dia.mainloop()

	"""
	Creates menu
	"""
	def menu(self):
		self.menu = Menu(self.dia)
		self.dia.config(menu=self.menu)

	def menuitems(self, menulabel, itemdict):
		submenu = Menu(self.menu)
		self.menu.add_cascade(label=menulabel, menu=submenu, font = "SansSerif, "+str(self.fs))
		itemlist = list(itemdict.keys())
		itemlist.sort()
		for entry in itemlist:
			valuetup = itemdict[entry]
			itemlabel = valuetup[0]
			if itemlabel == "Separator":
				submenu.add_separator()
			else:
				itemcommand = valuetup[1]
				submenu.add_command(label=itemlabel, command=itemcommand, font = "SansSerif, "+str(self.fs))

	def makestatusbar(self, sbtext):
		"""Creates the statusbar (at the bottom of the dialog)"""
		pad = 1
		self.statusbar = Label(self.dia, text="", bd=1, relief=SUNKEN, anchor=W)
		self.statusbar.config(text=sbtext)
		self.statusbar.pack(side="bottom", fill="x", pady=pad, padx=pad)
		return self.statusbar

	def makeframe(self):
		"""Creates frame"""
		self.frame = Frame(self.dia, borderwidth=2, bg = self.darkbgcolor)
		self.frame.pack(fill="both",expand=1)
		return self.frame

	def endframe(self):
		"""Destroys the frame"""
		self.frame.destroy()

	def ende(self):
		"""Destroys the dialog"""
		self.dia.destroy()

	"""
	General elements
	"""

	def button(self, buttontext, befehl, row_, column_):
		"""Generates a button"""
		b = Button(self.frame, text = buttontext, command = befehl)
		b.grid(row=row_, column=column_, padx=10)
		return b

	def label(self, ausgabetext, row_, column_, columnspan_=1):
		"""Creates a label field"""
		lb = Label(self.frame, text=ausgabetext, bg=self.brightcolor)
		lb.grid(row=row_, column=column_, columnspan=columnspan_)
		return lb

	def entryfield(self, row_, column_, width_=50, columnspan_=1):
		"""Creates an entry field"""
		e = Entry(self.frame, width=width_)
		e.grid(row=row_, column=column_, columnspan=columnspan_, padx=5, pady=10)
		return e

	def entrypasswd(self, row_, column_, width_=20, columnspan_=1):
		"""Creates an entry field, which displays * instead of the entried charakter"""
		epw = Entry(self.frame, width=width_, show="*")
		epw.grid(row=row_, column=column_, columnspan=columnspan_, padx=5, pady=10)
		return epw

	def listbox(self, liste_, width_, height_, row_, column_, columnspan_=1):
		"""Creates a listbox with scrollbar and multiple selections"""

		"""Creates a vertical scrollbar"""
		scb = Scrollbar(self.frame, orient="vertical")
		"""Creates a listbox"""
		listbox = Listbox(self.frame, width=width_, height=height_, selectmode = "multiple", yscrollcommand = scb.set)
		scb["command"] = listbox.yview
		for i in range(0,len(liste_)):
			listbox.insert("end", "  "+liste_[i])
		listbox.grid(row=row_, column=column_, padx=5)
		scb.grid(row=row_, column=column_+1, sticky=NS)
		return listbox

	def selecteditems(self, listbox):
		"""Realizes the selection(s) in a listbox"""
		# selecteditem=listbox.get("active").strip()
		selecteditems = []
		for element in listbox.curselection():
			selecteditems.append(listbox.get(element))
		return selecteditems

	def checkbox (self, row_, column_, text_, variable_, offvalue_=0, onvalue_=1):
		checkbox = Checkbutton(self.frame, text=text_, offvalue=offvalue_, onvalue=onvalue_, variable=variable_)
		checkbox.grid(row=row_, column=column_, padx=5, pady=10, sticky=W)
		return checkbox

	def minfo(self, mbtitle, mbtext, btype="ok"):
		sb = messagebox.showinfo(mbtitle, mbtext, type=btype)
		return sb

	def mwarning(self, mbtitle, mbtext, btype="ok"):
		sb = messagebox.showwarning(mbtitle, mbtext, type=btype)
		return sb

	def merror(self, mbtitle, mbtext, btype="ok"):
		sb = messagebox.showerror(mbtitle, mbtext, type=btype)
		return sb


class PDF2OCR (DialogMaker):

	def mainDialog (self):

		"""
		Variables
		-------------------
		for optionsframe
		"""
		self.sidecarflag = 0
		self.cleanflag = 0
		self.removebgflag = 0
		self.rotateflag = 0
		self.langflag = 0
		self.lang = ""
		self.addoptflag = 0
		self.addopt = ""
		self.options =""

		"""
		Creates dialog, menu and first frame.
		-------------------
		Creates the dialog
		"""
		self.makedia()
		self.diaminsize(diah=600, diav=400)
		self.geometry()

		"""Creates the menu"""
		self.menu()
		tasklabel = _("Menu")
		"""
		The key of the dictonary is the position of the menu item;
		the first entry of the value tupel is the label of the item
		and the second is the method, which will be excecuted by 
		clicking the item
		"""
		taskitemsdict = {
			1 : (_("Apply OCR to PDF"), self.appframe),
			2 : (_("Select options"), self.optionsframe),
			3 : (_("Program information"), self.programinfo),
			4 : (_("Licence"), self.licenceinfo),
			5 : ("Separator",),
			6 : (_("Quit"), self.doQuit)
			}
		self.menuitems(tasklabel, taskitemsdict)

		self.diatitle("ocrpdf-tk")
		self.makestatusbar(_("Start"))
		self.programinfo()
		self.mainloop()

	def appframe(self):
		try:
			self.endframe()
		except:
			pass

		self.makeframe()

		dplabel = self.label(_("Path:"), 0 ,0)
		dplabel.grid(sticky=W)
		self.initentry = self.entryfield(1, 0)
		self.initentry.grid(sticky=W)
		self.initentry.insert(0, self.homepath)
		workbutton = self.button(_("Next"), self.selectFile, 4, 2)
		ol=self.label("Selected options:", 5, 0)
		ol.grid(sticky=W)
		self.rl = self.label("", 7, 0, 3)
		self.rl.grid(sticky=W)
		self.showOptions()
		self.statusbar["text"] = _("Entry")

	def optionsframe(self):
		try:
			self.endframe()
		except:
			pass

		self.makeframe()

		hl=_("Select options for OCRmyPDF (see man ocrmypdf)")
		headline = self.label(hl, 0, 0, 4)
		headline.grid(sticky=W)

		self.wert0 = IntVar()
		self.wert1 = IntVar()
		self.wert2 = IntVar()
		self.wert3 = IntVar()
		self.wert4 = IntVar()
		self.wert5 = IntVar()

		checkbutton0 = self.checkbox (1, 0, _("Create text file"), self.wert0)
		checkbutton0["command"] = self.check0
		if self.sidecarflag == 1:
			checkbutton0.select()
		else:
			checkbutton0.deselect()

		checkbutton1 = self.checkbox (2, 0, _("Clean PDF"), self.wert1)
		checkbutton1["command"] = self.check0
		if self.cleanflag == 1:
			checkbutton1.select()
		else:
			checkbutton1.deselect()

		checkbutton2 = self.checkbox (3, 0, _("Remove background"), self.wert2)
		checkbutton2["command"] = self.check0
		if self.removebgflag == 1:
			checkbutton2.select()
		else:
			checkbutton2.deselect()

		checkbutton3 = self.checkbox (4, 0, _("Rotate pages"), self.wert3)
		checkbutton3["command"] = self.check0
		if self.rotateflag == 1:
			checkbutton3.select()
		else:
			checkbutton3.deselect()

		langlabel = self.label(_("Language to be recognized:"), 5, 0)
		langlabel.grid(sticky=W)

		self.checkbutton4 = self.checkbox (5, 2, _("Apply"), self.wert4)
		self.checkbutton4["command"] = self.check0
		if self.langflag == 1:
			self.checkbutton4.select()
		else:
			self.checkbutton4.deselect()

		self.langentry = self.entryfield(5, 1)
		self.langentry["width"] = 10
		self.langentry.grid(sticky=W)
		self.langentry.bind("<Button-1>", self.check1)
		lang = self.lang.replace('--language ', '')
		self.langentry.insert(0, lang)

		addoptlabel = self.label(_("Additional option(s):"), 6, 0)
		addoptlabel.grid(sticky=W)

		self.checkbutton5 = self.checkbox (6, 2, _("Apply"), self.wert5)
		self.checkbutton5["command"] = self.check0
		if self.addoptflag == 1:
			self.checkbutton5.select()
		else:
			self.checkbutton5.deselect()

		self.addoptentry = self.entryfield(6, 1)
		self.addoptentry["width"] = 50
		self.addoptentry.grid(sticky=W)
		self.addoptentry.bind("<Button-1>", self.check2)
		self.addoptentry.insert(0, self.addopt)

		self.rl = self.label("", 7, 0, 3)
		self.rl.grid(sticky=W)
		self.showOptions()

		self.statusbar["text"] = _("Select options")

	def check0 (self):
		if self.wert0.get() == 1:
			sidec = "--sidecar "
			self.sidecarflag = 1
		else:
			sidec = ""
			self.sidecarflag = 0

		if self.wert1.get() == 1:
			cl = "--clean "
			self.cleanflag = 1
		else:
			cl = ""
			self.cleanflag = 0

		if self.wert2.get() == 1:
			rmbg = "--remove-background "
			self.removebgflag = 1
		else:
			rmbg = ""
			self.removebgflag = 0

		if self.wert3.get() == 1:
			rotatepgs = "--rotate-pages "
			self.rotateflag = 1
		else:
			rotatepgs = ""
			self.rotateflag = 0

		if self.wert4.get() == 1:
			self.langflag = 1
			if self.langentry.get () == "":
				self.langentry.insert(0, "en")
				self.lang = "--language en"
			else:
				self.lang = "--language " + self.langentry.get ()
		else:
			self.langflag = 0
			self.lang = ""
			self.langentry.insert(0, "")

		if self.wert5.get() == 1:
			if self.addoptentry.get () != "":
				self.addopt = self.addoptentry.get ()
				self.addoptflag = 1
			else:
				self.checkbutton5.deselect()
				self.addoptflag = 0
		else:
			self.addoptflag = 0
			self.addopt = ""
			self.addoptentry.insert(0, "")


		self.options = sidec + cl +rmbg + rotatepgs
		self.showOptions()

	def showOptions (self):
		self.lang = self.lang.strip()
		self.addopt = " " + self.addopt.strip() + " "
		self.rl["text"] = self.options + self.lang + self.addopt

	def check1 (self, event):
		self.checkbutton4.deselect()
		self.lang = ""
		self.langflag = 0
		self.showOptions()

	def check2 (self, event):
		self.checkbutton5.deselect()
		self.addopt = ""
		self.addoptflag = 0
		self.showOptions()

	def programinfo(self):
		"""Show program information"""
		try:
			self.endframe()
		except:
			pass

		self.makeframe()
		lil = self.label(programname, 0, 0)
		lil["bg"] = self.darkbgcolor
		lil["font"] = "SansSerif, 12"
		self.statusbar["text"] = _("Program informations")

	def licenceinfo(self):
		"""Show licence"""
		try:
			self.endframe()
		except:
			pass

		self.makeframe()
		lil = self.label(licence, 0, 0)
		lil["bg"] = self.darkbgcolor
		lil["font"] = "SansSerif, 12"
		self.statusbar["text"] = _("Licence")

	def selectDir (self, ocrpath):
		"""Start"""
		self.ocrpath = ocrpath
		self.findUsersHome ()
		self.mainDialog ()

	def findUsersHome (self):
		"""Finds user's home directory on different operation systems"""
		if os.getenv('USERPROFILE'):
			self.homepath = os.getenv('USERPROFILE')
			self.os="W"
		elif os.getenv('HOME'):
			self.homepath = os.getenv('HOME')
			self.os="U"

	def selectFile (self):
		try:
			filedir = self.initentry.get()
		except:
			if self.os == "W":
				filedir = "C:\\"
			else:
				filedir = "/"

		"""Select file for working on"""
		self.file_path = filedialog.askopenfilename (initialdir = filedir, title = _("File selection"),filetypes = (("pdf files","*.pdf"), ("all files","*.*")))
		"""In case of cancel self.file_path is a tupel."""
		self.file_path = str(self.file_path)
		if self.file_path != "()":
			self.treatFile ()
			if sys.platform == "linux":
				self.statusbar["text"] = _("File will be displayed")
				if self.sidecarflag == 1: 
					self.displayText ()
				self.displayFile ()

	def treatFile (self):
		""" Apply OCRmyPDF - This is the core of the program! """
		self.newfile = ""
		RBC = ReplaceBC()
		if os.path.isfile(self.file_path):
			"""If a textfile should be created"""
			self.txtfile = self.file_path.replace('.pdf', '.txt')
			self.txtfile = RBC.replacechars(self.txtfile)
			options = self.options.replace('--sidecar', '--sidecar ' + self.txtfile)

			"""Build options string"""
			options = options + self.lang + self.addopt

			"""Rename output file"""
			self.newfile = self.file_path.replace('.pdf', '_ocr.pdf')
			self.newfile = RBC.replacechars(self.newfile)

			if not os.path.isfile(self.newfile):
				self.statusbar["text"] = _("Please wait ,,,")
				shcommand0 = self.ocrpath+" "+options+" "+self.file_path+" "+self.newfile
				try:
					os.system(shcommand0)
				except:
					e = sys.exc_info()
					self.merror(str(e[0]), _("Error: ")+str(e[1]))
			else:
				self.mwarning(_("Clash!"), self.newfile+_("\nalready exists!"))
				# self.dia.destroy()
		else:
			self.mwarning(_("Unable to do anything"), _("Please choose a file!"))

	def displayText (self):
		"""Displays OCR textfile"""
		if os.path.isfile("/usr/bin/xfce4-terminal"):
			os.system('xfce4-terminal -e "less -N '+self.txtfile+'" & ')
		else:
			os.system('xterm -e "less -N '+self.txtfile+'" & ')

	def displayFile (self):
		"""Shows new PDF file with atril"""
		pdfreader="/usr/bin/atril"
		if not os.path.isfile(pdfreader):
		   return 1
		if os.path.isfile(self.newfile):
			shcommand2 = pdfreader+" "+self.newfile+" & "
			os.system(shcommand2)

	def doQuit(self):
		"""
		ends the dialog
		"""
		self.dia.destroy()
		exit()


class ReplaceBC (object):

	"""Replace 'bad' characters in file names"""

	def replacechars(self, origstr):
		self.origstr = origstr
		self.replacebyunderscore()
		self.replacebynone()
		self.replaceumlauts()
		cleanstr = self.origstr
		return cleanstr

	"""Feel free to expand the lists """

	def replacebyunderscore(self):
		bclist = [",", " ", ";", "'", "\"", "%", "#", "!", ":", ]
		for element in bclist:
			self.origstr = self.origstr.replace (element, "_")

	def replacebynone(self):
		bclist = ["*", "?", "$", "\\", "`", "´", "|", "(", ")", "[",\
				  "]", "{", "}", "=", ]
		for element in bclist:
			self.origstr = self.origstr.replace (element, "")

	def replaceumlauts(self):
		self.origstr = self.origstr.replace ("Ä", "Ae")
		self.origstr = self.origstr.replace ("Ö", "Oe")
		self.origstr = self.origstr.replace ("Ü", "Ue")
		self.origstr = self.origstr.replace ("ä", "ae")
		self.origstr = self.origstr.replace ("ö", "oe")
		self.origstr = self.origstr.replace ("ü", "ue")
		self.origstr = self.origstr.replace ("ß", "ss")
		self.origstr = self.origstr.replace ("&", "et")
		self.origstr = self.origstr.replace ("§", "para")
		self.origstr = self.origstr.replace (">", "_gt_")
		self.origstr = self.origstr.replace ("<", "_lt_")


class TranslationIntegration(object):
	"""GNU-gettext for GNU/Linux"""

	def __init__(self, pname):
		self.moname = pname + ".mo"
		self.syslanguage()
		# check, whether it's a unix system installation
		_localepath = "/usr/share/locale/"

		_result = self.look4mo(_localepath)

		if _result == "xxx":
			self.syslanguage()
			_localepath = "/usr/local/share/locale/"
			_result = self.look4mo(_localepath)

		if _result == "xxx":
			self.syslanguage()
			_localepath = "locale/"
			_result = self.look4mo(_localepath)

		if not _result == "xxx":
			# chooses the right .mo file and installs translation
			trans = gettext.translation(pname, _localepath, [self.language])
			trans.install()
		else:
			# Translating the following strings is absurd!
			serror = "Bad package! No "+self.moname+" file found!"
			from tkinter import messagebox
			messagebox.showerror('Error!', serror)
			sys.exit(1)

	def syslanguage(self):
		"""which is the language of the system (GNU/Linux)?"""
		_loclang = locale.getdefaultlocale()
		self.language = _loclang[0]
		if self.language == None:
			self.language = "en"
		return self.language

	def look4mo(self, localepath):
		"""choose only a language which a .mo file exists for,
		else use english or return 'xxx' """
		_lpath = localepath + self.language + "/LC_MESSAGES/" + self.moname

		if not os.path.exists(_lpath):

			_newlang = self.dosplit(self.language, "_")
			_lpath = localepath + _newlang + "/LC_MESSAGES/"+self.moname

			if os.path.exists(_lpath):
				self.language = _newlang
			else:
				_newlang = self.dosplit(self.language, "@")
				_lpath = localepath + _newlang + "/LC_MESSAGES/" + self.moname

				if os.path.exists(_lpath):
					self.language = _newlang
				else:
					_lpath = localepath + "en/LC_MESSAGES/" + self.moname

					if os.path.exists(_lpath):
						self.language = "en"
					else:
						self.language = "xxx"

		return self.language

	def dosplit(self, language, splitter):

		"""
			splits language descriptor

			>>> gnugettext.dosplit(gnugettext, "de_DE", "_")
			'de'

			>>> gnugettext.dosplit(gnugettext, "be@latin", "@")
			'be'
		"""

		_newlang = language.split(splitter)
		_newlang = _newlang[0]

		return _newlang


class ConfigFileRW(object):

	def __init__ (self):

		self.confp = configparser.ConfigParser()
		
	def pathConfigFile(self, sconfigfile):	
		"""The configfile must be in the users directory"""
		config_path = None
		if os.getenv('USERPROFILE'):
			config_path = os.getenv('USERPROFILE')
		elif os.getenv('HOME'):
			config_path = os.getenv('HOME')
			"""Configfile is hidden in unixoid systems"""
			sconfigfile = "."+sconfigfile
		"""Building the path to the configfile"""
		if config_path:
			self.sconfigfile = os.path.join(config_path, sconfigfile)
		else:
			self.sconfigfile = sconfigfile
		return self.sconfigfile

	def readConfigFile(self, ssection, lprops):
		"""Reads the values of given properties
		   of a given section  of the given configfile
		   and returns a property-value-dictionary"""

		conf_dict = {}
		if self.sconfigfile:
			self.confp.read(self.sconfigfile)
			for element in  lprops:
				try:
					probs_str = self.confp.get(ssection, element)
				except:
					probs_str = ""
				conf_dict.update({element : probs_str})
		else:
			"""If there is not configfile,
			   all values are empty strings"""
			probs_str = ""
			for element in  lprops:
				probs_dict.update({element : probs_str})

		return conf_dict

	def readTheWholeConfigFile(self):
		"""reads all sections, properties and values
		   of a given configfile and returns a
		   dictionary, which properties are the sections
		   and values are property-value-dictionaries"""

		if self.sconfigfile:
			self.confp.read(self.sconfigfile)
			sect_prop_dict = {}
			conf_sections_list = self.confp.sections()
			for element in conf_sections_list:
				prop_list = []
				for prop in self.confp[element]:
					prop_list.append(prop)
				conf_dict = self.readConfigFile(element, prop_list)
				sect_prop_dict.update({element : conf_dict})
			return sect_prop_dict

	def writeConfigFile(self, ssection, dprops):
		"""Writes a section in configfile with properties and
		   values"""

		self.confp[ssection] = dprops
		
		with open(self.sconfigfile, 'w') as configfile:
			self.confp.write(configfile)

	def updateSection(self, ssection, dprops):
		"""Updates or appends a section in given configfile with properties and
		   values"""

		if self.sconfigfile:
			sect_prop_dict = self.readTheWholeConfigFile()
			sect_prop_dict.update({ssection :  dprops})

			for key in sect_prop_dict:
				self.writeConfigFile(key, sect_prop_dict[key])
		else:
			self.writeConfigFile(ssection, dprops)


def main():
	
	"""Gettext Integration"""
	packname = progname.replace(".py", "")
	GtI = TranslationIntegration(packname)

	"""Path zu OCRmyPDF"""
	if os.path.isfile('/usr/bin/ocrmypdf'):
		ocrPath='/usr/bin/ocrmypdf'
	else:
		CF = ConfigFileRW()
		cnffile = CF.pathConfigFile('ocrmypdfpath')
		if os.path.isfile(cnffile):
			ocrPathdict = CF.readConfigFile("PATH", ["path2ocrmypdf"])
			ocrPath = ocrPathdict["path2ocrmypdf"]
			if not os.path.isfile(ocrPath):
				os.remove(cnffile)
				ocrPath = filedialog.askopenfilename (initialdir = filedir, title = _("Find OCRmyPDF"))
				oCF.writeConfigFile('PATH', {"path2ocrmypdf" : ocrPath})
		else:
			if os.getenv('USERPROFILE'):
				filedir = "C:\\"
			elif os.getenv('HOME'):
				filedir = "/"
			ocrPath = filedialog.askopenfilename (initialdir = filedir, title = _("Find OCRmyPDF"))
			CF.writeConfigFile('PATH', {"path2ocrmypdf" : ocrPath})

	"""Start application"""
	App = PDF2OCR()
	"""Check, whether OCRmyPDF is at the right place"""
	if not os.path.isfile(ocrPath):
		App.mwarning(_("Path: ")+ocrPath, _("OCRmyPDF not found!"))
		exit()

	App.selectDir(ocrPath)
	return 0

if __name__ == '__main__':
	main()
